@@ -29,6 +29,7 @@
 #include <qdir.h>
 #include <qpointarray.h>
 #include <qmessagebox.h>
+#include <algorithm>
 #include <cstdlib>
 #include <cmath>
 #include <unistd.h>
@@ -105,13 +106,16 @@ QByteArray ComputeMD5Sum(QByteArray *in);
 char *toHex( uchar u );
 QString String2Hex(QString *in, bool lang = true);
 QString CompressStr(QString *in);
+QImage scaleImage(const QImage& src, int width, int height);
 QString ImageToTxt(QImage *im);
 QString ImageToCMYK(QImage *im);
 QString ImageToCMYK_PS(QImage *im, int pl, bool pre);
+bool isProgressive(QString fn);
 void Convert2JPG(QString fn, QImage *image, int Quality, bool isCMYK);
 QString MaskToTxt(QImage *im, bool PDF = true);
 QString MaskToTxt14(QImage *im);
 void Level2Layer(ScribusDoc *doc, struct Layer *ll, int Level);
+int Layer2Level(ScribusDoc *currentDoc, int LayerNr);
 void BezierPoints(QPointArray *ar, QPoint n1, QPoint n2, QPoint n3, QPoint n4);
 double xy2Deg(double x, double y);
 QPointArray FlattenPath(FPointArray ina, QValueList<uint> &Segs);
@@ -122,6 +126,8 @@ uint getDouble(QString in, bool raw);
 bool loadText(QString nam, QString *Buffer);
 double Cwidth(ScribusDoc *doc, QString name, QString ch, int Siz, QString ch2 = " ");
 double RealCWidth(ScribusDoc *doc, QString name, QString ch, int Siz);
+double RealCHeight(ScribusDoc *currentDoc, QString name, QString ch, int Size);
+double RealCAscent(ScribusDoc *currentDoc, QString name, QString ch, int Size);
 double QStodouble(QString in);
 int QStoInt(QString in);
 QString GetAttr(QDomElement *el, QString at, QString def="0");
@@ -133,7 +139,6 @@ QImage LoadPictCol(QString fn, QString Prof, bool UseEmbedded, bool *realCMYK);
 QImage ProofPict(QImage *Im, QString Prof, int Rend);
 #endif
 QImage ProofImage(QImage *Im);
-int System(const QStringList & args);
 int callGS(const QStringList & args_in);
 int copyFile(QString source, QString target);
 int moveFile(QString source, QString target);
@@ -238,34 +243,6 @@ QImage ProofImage(QImage *Im)
 #endif
 }
 
-/******************************************************************
- * Function System()
- *  
- * Create a new process via QProcess and wait until finished.
- * return the process exit code.
- *
- ******************************************************************/
-
-int System(const QStringList & args)
-{
-	QProcess *proc = new QProcess(NULL);
-	proc->setArguments(args);
-	if ( !proc->start() )
-	{
-		delete proc;
-		return 1;
-	}
-	/* start was OK */
-	while(proc->isRunning())
-	{
-		/* wait a little bit */
-		usleep(5000);
-	}
-	int ex = proc->exitStatus();
-	delete proc;
-	return ex;
-}
-
 /******************************************************************
  * Function callGS()
  *   build the complete list of arguments for the call of our
@@ -278,31 +255,34 @@ int System(const QStringList & args)
  
 int callGS(const QStringList & args_in)
 {
-	QStringList args;
-
-	/* these parameters are always the same */
-	args.append(ScApp->Prefs.gs_exe);
-	args.append("-q");
-	args.append("-dNOPAUSE");
+	QString cmd1 = ScApp->Prefs.gs_exe;
+	cmd1 += " -q -dNOPAUSE";
 	if (ScApp->HavePngAlpha != 0)
-		args.append("-sDEVICE=png16m");
+		cmd1 += " -sDEVICE=png16m";
 	else
-		args.append("-sDEVICE=pngalpha");
+		cmd1 += " -sDEVICE=pngalpha";
 	if (ScApp->Prefs.gs_antiText)
-		args.append("-dTextAlphaBits=4");
+		cmd1 += " -dTextAlphaBits=4";
 	if (ScApp->Prefs.gs_antiGraph)
-		args.append("-dGraphicsAlphaBits=4");
-	
-	/* insert specific arguments */
-	QStringList p;
-	p = args_in;
-	args += args_in;
-	/* insert last unspecific arguments */
-	args.append("-c");
-	args.append("showpage");
-	args.append("-c");
-	args.append("quit");
-        return System(args);
+		cmd1 += " -dGraphicsAlphaBits=4";
+	// Add -sFONTPATH to tell gs where any additional font paths are
+	QStringList extraFonts;
+	QFile fx(ScApp->PrefsPfad+"/scribusfont.rc");
+	if (fx.open(IO_ReadOnly))
+	{
+		QTextStream tsx(&fx);
+		QString extraPath = tsx.read();
+		fx.close();
+		extraFonts = QStringList::split("\n", extraPath);
+	}
+	if (extraFonts.count() >= 1)
+		cmd1 += QString(" -sFONTPATH='%1'").arg(extraFonts[0]);
+	for (int i = 1; i < static_cast<int>(extraFonts.count()); ++i)
+		cmd1 += QString(":'%1'").arg(extraFonts[i]);
+	// then combine the cmdline and run gs
+	QString extArgs = args_in.join(" ");
+	cmd1 += " " + extArgs + " -c showpage -c quit";
+	return system(cmd1.local8Bit());
 }
 
 int copyFile(QString source, QString target)
@@ -340,6 +320,203 @@ int moveFile(QString source, QString target)
 	return 0;
 }
 
+QImage scaleImage(const QImage& src, int width, int height)
+{
+	QImage dst;
+	dst.create(width, height,32);
+	QRgb* xelrow = 0;
+	QRgb* tempxelrow = 0;
+	register QRgb* xP;
+	register QRgb* nxP;
+	int rows, cols, rowsread, newrows, newcols;
+	register int row, col, needtoreadrow;
+	const uchar maxval = 255;
+	double xscale, yscale;
+	long sxscale, syscale;
+	register long fracrowtofill, fracrowleft;
+	long* as;
+	long* rs;
+	long* gs;
+	long* bs;
+	int rowswritten = 0;
+	cols = src.width();
+	rows = src.height();
+	newcols = dst.width();
+	newrows = dst.height();
+	long SCALE;
+	long HALFSCALE;
+	if (cols > 4096)
+	{
+		SCALE = 4096;
+		HALFSCALE = 2048;
+	}
+	else
+	{
+		int fac = 4096;
+		while (cols * fac > 4096)
+		{
+			fac /= 2;
+		}
+		SCALE = fac * cols;
+		HALFSCALE = fac * cols / 2;
+	}
+	xscale = (double) newcols / (double) cols;
+	yscale = (double) newrows / (double) rows;
+	sxscale = (long)(xscale * SCALE);
+	syscale = (long)(yscale * SCALE);
+	if ( newrows != rows )	/* shortcut Y scaling if possible */
+		tempxelrow = new QRgb[cols];
+	as = new long[cols];
+	rs = new long[cols];
+	gs = new long[cols];
+	bs = new long[cols];
+	rowsread = 0;
+	fracrowleft = syscale;
+	needtoreadrow = 1;
+	for ( col = 0; col < cols; ++col )
+		rs[col] = gs[col] =  as[col] = bs[col] = HALFSCALE;
+	fracrowtofill = SCALE;
+	for ( row = 0; row < newrows; ++row )
+	{
+		if ( newrows == rows )
+			tempxelrow = xelrow = (QRgb*)src.scanLine(rowsread++);
+		else
+		{
+			while ( fracrowleft < fracrowtofill )
+			{
+				if ( needtoreadrow && rowsread < rows )
+					xelrow = (QRgb*)src.scanLine(rowsread++);
+				for ( col = 0, xP = xelrow; col < cols; ++col, ++xP )
+				{
+					as[col] += fracrowleft * qAlpha( *xP );
+					rs[col] += fracrowleft * qRed( *xP );
+					gs[col] += fracrowleft * qGreen( *xP );
+					bs[col] += fracrowleft * qBlue( *xP );
+				}
+				fracrowtofill -= fracrowleft;
+				fracrowleft = syscale;
+				needtoreadrow = 1;
+			}
+			if ( needtoreadrow && rowsread < rows )
+			{
+				xelrow = (QRgb*)src.scanLine(rowsread++);
+				needtoreadrow = 0;
+			}
+			register long a=0;
+			for ( col = 0, xP = xelrow, nxP = tempxelrow; col < cols; ++col, ++xP, ++nxP )
+			{
+				register long r, g, b;
+				a = as[col] + fracrowtofill * qAlpha( *xP );
+				r = rs[col] + fracrowtofill * qRed( *xP );
+				g = gs[col] + fracrowtofill * qGreen( *xP );
+				b = bs[col] + fracrowtofill * qBlue( *xP );
+				r /= SCALE;
+				if ( r > maxval ) r = maxval;
+				g /= SCALE;
+				if ( g > maxval ) g = maxval;
+				b /= SCALE;
+				if ( b > maxval ) b = maxval;
+				a /= SCALE;
+				if ( a > maxval ) a = maxval;
+				*nxP = qRgba( (int)r, (int)g, (int)b , (int)a);
+				rs[col] = as[col] = gs[col] = bs[col] = HALFSCALE;
+			}
+			fracrowleft -= fracrowtofill;
+			if ( fracrowleft == 0 )
+			{
+				fracrowleft = syscale;
+				needtoreadrow = 1;
+			}
+			fracrowtofill = SCALE;
+		}
+		if ( newcols == cols )
+			memcpy(dst.scanLine(rowswritten++), tempxelrow, newcols*4);
+		else
+		{
+			register long a, r, g, b;
+			register long fraccoltofill, fraccolleft = 0;
+			register int needcol;
+			nxP = (QRgb*)dst.scanLine(rowswritten++);
+			fraccoltofill = SCALE;
+			a = r = g = b = HALFSCALE;
+			needcol = 0;
+			for ( col = 0, xP = tempxelrow; col < cols; ++col, ++xP )
+			{
+				fraccolleft = sxscale;
+				while ( fraccolleft >= fraccoltofill )
+				{
+					if ( needcol )
+					{
+						++nxP;
+						a = r = g = b = HALFSCALE;
+					}
+					a += fraccoltofill * qAlpha( *xP );
+					r += fraccoltofill * qRed( *xP );
+					g += fraccoltofill * qGreen( *xP );
+					b += fraccoltofill * qBlue( *xP );
+					r /= SCALE;
+					if ( r > maxval ) r = maxval;
+					g /= SCALE;
+					if ( g > maxval ) g = maxval;
+					b /= SCALE;
+					if ( b > maxval ) b = maxval;
+					a /= SCALE;
+					if ( a > maxval ) a = maxval;
+					*nxP = qRgba( (int)r, (int)g, (int)b, (int)a );
+					fraccolleft -= fraccoltofill;
+					fraccoltofill = SCALE;
+					needcol = 1;
+				}
+				if ( fraccolleft > 0 )
+				{
+					if ( needcol )
+					{
+						++nxP;
+						a = r = g = b = HALFSCALE;
+						needcol = 0;
+					}
+					a += fraccolleft * qAlpha( *xP );
+					r += fraccolleft * qRed( *xP );
+					g += fraccolleft * qGreen( *xP );
+					b += fraccolleft * qBlue( *xP );
+					fraccoltofill -= fraccolleft;
+				}
+			}
+			if ( fraccoltofill > 0 )
+			{
+				--xP;
+				a += fraccolleft * qAlpha( *xP );
+				r += fraccoltofill * qRed( *xP );
+				g += fraccoltofill * qGreen( *xP );
+				b += fraccoltofill * qBlue( *xP );
+			}
+			if ( ! needcol )
+			{
+				r /= SCALE;
+				if ( r > maxval ) r = maxval;
+				g /= SCALE;
+				if ( g > maxval ) g = maxval;
+				b /= SCALE;
+				if ( b > maxval ) b = maxval;
+				a /= SCALE;
+				if ( a > maxval ) a = maxval;
+				*nxP = qRgba( (int)r, (int)g, (int)b, (int)a );
+			}
+		}
+	}
+	if ( newrows != rows && tempxelrow )// Robust, tempxelrow might be 0 1 day
+		delete [] tempxelrow;
+	if ( as )				// Avoid purify complaint
+		delete [] as;
+	if ( rs )				// Robust, rs might be 0 one day
+		delete [] rs;
+	if ( gs )				// Robust, gs might be 0 one day
+		delete [] gs;
+	if ( bs )				// Robust, bs might be 0 one day
+		delete [] bs;
+	return dst;
+}
+
 QPixmap LoadPDF(QString fn, int Seite, int Size, int *w, int *h)
 {
 	QString tmp, cmd1, cmd2;
@@ -352,7 +529,7 @@ QPixmap LoadPDF(QString fn, int Seite, int Size, int *w, int *h)
 	args.append("-sOutputFile="+tmpFile);
 	args.append("-dFirstPage="+tmp);
 	args.append("-dLastPage="+tmp);
-	args.append(fn);
+	args.append("\""+fn+"\"");
 	ret = callGS(args);
 	if (ret == 0)
 	{
@@ -396,7 +573,7 @@ QImage LoadPict(QString fn, bool *gray)
 		args.append("-sOutputFile="+tmpFile);
 		args.append("-dFirstPage=1");
 		args.append("-dLastPage=1");
-		args.append(fn);
+		args.append("\""+fn+"\"");
 		ret = callGS(args);
 		if (ret == 0)
 		{
@@ -447,7 +624,7 @@ QImage LoadPict(QString fn, bool *gray)
 				args.append("-r72");
 				args.append("-sOutputFile="+tmpFile);
 				args.append("-g"+tmp.setNum(qRound(b))+"x"+tmp2.setNum(qRound(h)));
-				args.append(fn);
+				args.append("\""+fn+"\"");
 				ret = callGS(args);
 				if (ret == 0)
 				{
@@ -481,7 +658,7 @@ QImage LoadPict(QString fn, bool *gray)
 	{
 		QImage img;
 		QImage inI2;
-		TIFF* tif = TIFFOpen(fn, "r");
+		TIFF* tif = TIFFOpen(fn.local8Bit(), "r");
 		if(tif)
 		{
 			unsigned width, height,size;
@@ -559,7 +736,7 @@ QImage LoadPictCol(QString fn, QString Prof, bool UseEmbedded, bool *realCMYK)
 		args.append("-sOutputFile="+tmpFile);
 		args.append("-dFirstPage=1");
 		args.append("-dLastPage=1");
-		args.append(fn);
+		args.append("\""+fn+"\"");
 		ret = callGS(args);
 		if (ret == 0)
 		{
@@ -611,7 +788,7 @@ QImage LoadPictCol(QString fn, QString Prof, bool UseEmbedded, bool *realCMYK)
 					args.append("-r72");
 					args.append("-sOutputFile="+tmpFile);
 					args.append("-g"+tmp.setNum(qRound(b))+"x"+tmp2.setNum(qRound(h)));
-					args.append(fn);
+					args.append("\""+fn+"\"");
 					ret = callGS(args);
 					if (ret == 0)
 					{
@@ -646,7 +823,7 @@ QImage LoadPictCol(QString fn, QString Prof, bool UseEmbedded, bool *realCMYK)
 		{
 			QImage img;
 			QImage inI2;
-			TIFF* tif = TIFFOpen(fn, "r");
+			TIFF* tif = TIFFOpen(fn.local8Bit(), "r");
 			if(tif)
 			{
     			DWORD EmbedLen = 0;
@@ -886,6 +1063,42 @@ double RealCWidth(ScribusDoc *doc, QString name, QString ch, int Siz)
 		return static_cast<double>(Siz / 10.0);
 }
 
+double RealCHeight(ScribusDoc *currentDoc, QString name, QString ch, int Size)
+{
+	double w;
+	uint c1 = ch.at(0).unicode();
+	FT_Face      face;
+	Foi* fo = (*currentDoc->AllFonts)[name];
+	if (fo->CharWidth.contains(c1))
+	{
+		face = currentDoc->FFonts[name];
+		uint cl = FT_Get_Char_Index(face, c1);
+		FT_Load_Glyph(face, cl, FT_LOAD_NO_SCALE | FT_LOAD_NO_BITMAP );
+		w = face->glyph->metrics.height / fo->uniEM * (Size / 10.0);
+		return w;
+	}
+	else
+		return static_cast<double>(Size / 10.0);
+}
+
+double RealCAscent(ScribusDoc *currentDoc, QString name, QString ch, int Size)
+{
+	double w;
+	uint c1 = ch.at(0).unicode();
+	FT_Face      face;
+	Foi* fo = (*currentDoc->AllFonts)[name];
+	if (fo->CharWidth.contains(c1))
+	{
+		face = currentDoc->FFonts[name];
+		uint cl = FT_Get_Char_Index(face, c1);
+		FT_Load_Glyph(face, cl, FT_LOAD_NO_SCALE | FT_LOAD_NO_BITMAP );
+		w = face->glyph->metrics.horiBearingY / fo->uniEM * (Size / 10.0);
+		return w;
+	}
+	else
+		return static_cast<double>(Size / 10.0);
+}
+
 QPointArray RegularPolygon(double w, double h, uint c, bool star, double factor, double rota)
 {
 	uint cx = star ? c * 2 : c;
@@ -1000,6 +1213,16 @@ void Level2Layer(ScribusDoc *doc, struct Layer *ll, int Level)
 	}
 }
 
+int Layer2Level(ScribusDoc *currentDoc, int LayerNr)
+{
+	for (uint la2 = 0; la2 < currentDoc->Layers.count(); ++la2)
+	{
+		if (currentDoc->Layers[la2].LNr == LayerNr)
+			return currentDoc->Layers[la2].Level;
+	}
+	return 0;
+}
+
 QString ImageToTxt(QImage *im)
 {
 	int h = im->height();
@@ -1182,6 +1405,33 @@ static void my_error_exit (j_common_ptr cinfo)
   longjmp (myerr->setjmp_buffer, 1);
 }
 
+bool isProgressive(QString fn)
+{
+	struct jpeg_decompress_struct cinfo;
+	struct my_error_mgr         jerr;
+	FILE     *infile;
+	bool ret = false;
+	cinfo.err = jpeg_std_error (&jerr.pub);
+	jerr.pub.error_exit = my_error_exit;
+	infile = NULL;
+	if (setjmp (jerr.setjmp_buffer))
+	{
+		jpeg_destroy_decompress (&cinfo);
+		if (infile)
+			fclose (infile);
+		return ret;
+	}
+	jpeg_create_decompress (&cinfo);
+	if ((infile = fopen (fn.local8Bit(), "rb")) == NULL)
+		return ret;
+	jpeg_stdio_src(&cinfo, infile);
+	jpeg_read_header(&cinfo, TRUE);
+	ret = jpeg_has_multiple_scans(&cinfo);
+	fclose (infile);
+	jpeg_destroy_decompress (&cinfo);
+	return ret;
+}
+
 void Convert2JPG(QString fn, QImage *image, int Quality, bool isCMYK)
 {
 	struct jpeg_compress_struct cinfo;
@@ -1200,7 +1450,7 @@ void Convert2JPG(QString fn, QImage *image, int Quality, bool isCMYK)
 		return;
 	}
 	jpeg_create_compress (&cinfo);
-	if ((outfile = fopen (fn, "wb")) == NULL)
+	if ((outfile = fopen (fn.local8Bit(), "wb")) == NULL)
 		return;
 	jpeg_stdio_dest (&cinfo, outfile);
 	cinfo.image_width  = image->width();
@@ -1381,7 +1631,7 @@ int setBestEncoding(FT_Face face)
 			break;
 		}
 	}
-	if (countUniCode > 255)
+	if (countUniCode >= face->num_glyphs-1)
 	{
 		FT_Set_Charmap(face, face->charmaps[chmapUniCode]);
 		retVal = 0;
